package org.qatools.validator;


import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import org.apache.log4j.Logger;
import org.qatools.info.Info;


/**
 * <pre>
 * Scott Wilson presents
 * QATools.org
 * We Help You Lower Your QA Costs
 * Specializing in Functional & Integration Test Automation using Maven Java Junit TestNG
 * 
 * 
 * Validator extends Junit by capturing and reporting all failures at once.
 * This sppeds your QA effort by reducing the number of iterations.
 * 
 * Junit pukes at the first failed assertion and also does not enforce useful error messages
 * 			assertTrue(false);		<=== FAILS IMMEDIATELY
 * 			assertEquals(1, 2)		
 * 			assertNotNull(null);	
 * 
 * Validator example - the API forces you to enter useful error messages
 * 			Validator v = new Validator();
 * 			v.assertTrue("Foo msg", false);
 * 			v.assertEquals("Bar msg", 1, 2);
 * 			v.assertNotNull("Apple", null);
 * 			v.throwExceptionIfErrors();		<=== Captures ALL errors not just the first one!
 * 
 * </pre>
 * 
 * @author Scott Wilson - QATools.org
 */
public class Validator {
	private static Info info = new Info();	// Automatically print blog info... 
											// unless it's deactivated prior to loading this class with QATools.disableIn()
									
    private static Logger logger = Logger.getLogger(Validator.class);

	
    // ------------------------------------------------------
    // ----- static initialization
    // ------------------------------------------------------

    static {
    	//addPrintFormat(PrintFormat.BR);				// error messages contain <br> for html reports
    	addPrintFormat(PrintFormat.NEWLINE);		// error messages contain \n
        addPrintFormat(PrintFormat.NUMBER_ERRORS);  // errors are numbered
        //addPrintFormat(PrintFormat.HASH_MARKS);  	// add headers to separate errors & debugs with hash marks
    	
    	
        //
        // control how much error & debug info shows up in the surefire report
        // allows you to decrease how much info shows up in your build systems
    	setErrorOption(ErrorOption.SUREFIRE_FULL);	// ErrorOption enum allows you to control how much info shows up in the surefire report
    	setDebugOption(DebugOption.EVERYWHERE);		// DebugOption enum allows you to control how much debug info shows up in the surefire report
    }
	
    // ------------------------------------------------------
    // ----- class variables
    // ------------------------------------------------------

    //errors
    private List<String> errors = null;
    
    //debug
    private List<String> debugs = null;
    
    // print formats
	private static List<PrintFormat> sPrintFormats;

    // error options - formatting errors
	private static ErrorOption sErrorOption;
	
    // debug options - formatting debug messages
	private static DebugOption sDebugOption;
	
	//
	// String Processing Options
	private static List<StringCleaningOption> sStringCleaningOptions;	// trim, remove dup spaces, null equivalence
	private static List<String> sNullEquivalenceStrings;					// customize null = "", " ", "null", etc - must be used with sStringProcessingOptions.NULL_EQUIVALENCE
	private static boolean sStringCleaningForReportFlag = true;	//so reports aren't cluttered by really long strings
		//TODO - could change this to concatLongStringsForReport(int length)

	//
	// Global static debug details
	private static List<String> sDebugs = null;
	private static boolean sStaticDebuggingFlag = false;    // allows global debug messaging - must turn it on to use it
    private static boolean sStaticDebuggingAutoCleanOnMerge = true;    // automatically merge & clean static debugs
																		// client must resetStaticProperties if this property is false
    //TODO automatically associate static debugs to the TestCase rather than the Validator
    
    // ------------------------------------------------------
    // ----- constructors
    // ------------------------------------------------------

	public Validator() {
        //do nothing constructor
    }
    

    // ------------------------------------------------------
    // ----- errors
    // ------------------------------------------------------
    
	
    /**
     * Adds an error to the error list
     * 
     * @param error
     */
    public void addError(String error) {
    	if (errors == null) {
    		errors = new LinkedList<String>();
    	}
    	errors.add(error);
    }
	
    
    /**
     * Check if there are any errors
     * 
     * @return
     */
    public boolean doErrorsExist() {
    	if (errors==null || errors.size()==0) {
    		return false;
    	}
    	return true;
    }

    /**
     * Returns the list of errors
     * 
     * @return
     */
    public List<String> getErrors() {
        return errors;
    }

    
    // ----------------------------------------------------------------------------------------
    // ----- debugs - there are class & global static debugs
    // ----------------------------------------------------------------------------------------

    /**
     * Static debugging is OFF BY DEFAULT because some legacy frameworks using Validator do not exclusively use Validator. 
     * Test cases running only Junit might invoke utilities that do static debugging but these will not be cleaned up by test cases not invoking Validator assertions..
     * You must enable this feature and take responsibility for invoking Validator.resetStaticProperties() in your tearDown so the Junit ONLY test cases don't add out of scope static debug messages.
     */
    public static void disableStaticDebugging() {
        sStaticDebuggingFlag = false;        
    }
    /**
     * Static debugging is OFF BY DEFAULT because some legacy frameworks using Validator do not exclusively use Validator. 
     * Test cases running only Junit might invoke utilities that do static debugging but these will not be cleaned up by test cases not invoking Validator assertions..
     * You must enable this feature and take responsibility for invoking Validator.resetStaticProperties() in your tearDown so the Junit ONLY test cases don't add out of scope static debug messages.
     */
    public static void enableStaticDebugging() {
        sStaticDebuggingFlag = true;  
    }
    
    
    
    /**
     * Auto cleaning static debugging is enabled by default so debug messages are cleaned & do not pile up.
     * But this is not always desierable if you are using multiple validators. If the first validator merges & resets static debug message but has no errors,
     * then the second validator will lose the static debug messagets.
     * 
     * By dsiabling autoCleanOnMerge you can keep static debug messages after you assertions, but you must take responsibility
     * to lean them up manually be invoking Validator.resetStaticProperties()
     */
    public static void disableStaticDebuggingAutoCleanOnMerge() {
        sStaticDebuggingAutoCleanOnMerge = false;
    }
    /**
     * Auto cleaning static debugging is enabled by default so debug messages are cleaned & do not pile up.
     * But this is not always desierable if you are using multiple validators. If the first validator merges & resets static debug message but has no errors,
     * then the second validator will lose the static debug messagets.
     * 
     * By dsiabling autoCleanOnMerge you can keep static debug messages after you assertions, but you must take responsibility
     * to lean them up manually be invoking Validator.resetStaticProperties()
     */
    public static void enableStaticDebuggingAutoCleanOnMerge() {
        sStaticDebuggingAutoCleanOnMerge = true;        
    }
    
    
    /**
     * Returns the list of debug messages - CLASS & STATIC
     * 
     * @return
     */
    public List<String> getDebugs() {
        if (debugs == null && sDebugs == null) {
            return null;
            
        } else if (debugs != null && sDebugs != null) {
            List<String> list = new LinkedList<String>();
            list.addAll (debugs);
            list.addAll (sDebugs);
            return list;
            
        } else if (debugs != null) {
            return debugs;
        }
        return sDebugs;
    }
//TODO not sure this is needed since there is also a mergeDebugs()
    
    
    
    /**
     * Adds debug info to the debug list - CLASS INSTANCE
     * 
     * @param error
     */
    public void addDebug(String info) {
        if (debugs == null) {
            debugs = new LinkedList<String>();
        }
        debugs.add(info);
    }
    public void addDebug(String key, Object value) {
        addDebug(key + " = " + value);
    }

    
    /**
     * Adds an info to the debug list - STATIC GLOBAL LIST
     * 
     * Static debugging must be enabled otherwise it is not added to the list
     * 
     * @param error
     */
    public static void addStaticDebug(String info) {
        if (sStaticDebuggingFlag == false) {
            return;
        }
        //TODO - change the implementation to add it to the TestCase instead of the Validator
        //		but this might not work if the setUP is adding static debugs because the testcase is
        //		not known yet
        // maybe the getErrMsgIfExists() should move debugs to the testcase
        
        if (sDebugs == null) {
            sDebugs = new LinkedList<String>();
        }
        sDebugs.add(info);
    }
    public static void addStaticDebug(String key, Object value) {
        addStaticDebug(key + " = " + value);
    }

    
    /**
     * Check if there are any debug messages
     * 
     * @return
     */
    public boolean doDebugsExist() {
        if ( (debugs != null && 0 < debugs.size ()) || (sDebugs != null && 0 < sDebugs.size ()) ) {
            return true;
        }
        return false;
    }

    /**
     * Merge CLASS & STATIC debugs
     * 
     * Reset Static global debugs if sStaticDebuggingAutoCleanOnMerge == true
     */
    private void mergeDebugs() {
        if (sDebugs != null && 0<sDebugs.size ()) {
            if (debugs==null) {
                debugs = new LinkedList<String>();
            }
            debugs.addAll (sDebugs);
            
            //
            // automatically clean up static debugs
            // you might want to disable this if you're using multiple validator methods
            // where the first one passes but the second one fails, you dont want the first
            // validator to clean up the static debugs
            if (sStaticDebuggingAutoCleanOnMerge == true) {
                sDebugs = null;
            }
        }
    }
    
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

    /**
     * You can control long strings in the DetailReport by disabling it
     */
    public static void disableStringProcessingForReportFlag() {
    	sStringCleaningForReportFlag = false;
    }
    
    /**
     * You can control long strings in the DetailReport by disabling it
     */
    public static void enableStringProcessingForReportFlag() {
    	sStringCleaningForReportFlag = true;
    }
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------
    
    /**
     * 
     */
    public void junitAssertNoErrors() throws Exception {
    	throwExceptionIfErrors();
    }
    
    /**
     * 
     */
    public void testNGAssertNoErrors() throws Exception {
    	throwExceptionIfErrors();
    }

    
    /**
     * If any errors where captured, throw an Exception
     */
    public void throwExceptionIfErrors() throws Exception {
//    	DetailReport.resetAssertFlag();
        String msg = getErrorMsgIfExists();
        if (msg != null) {
            throw new Exception(msg);
        }
    }


    /**
     * If any errors where captured, throw an Exception
     */
    public void throwRuntimeExceptionIfErrors() {
//    	DetailReport.resetAssertFlag();
        String msg = getErrorMsgIfExists();
		if (msg != null) {
			throw new RuntimeException(msg);
		}
    }
    
    /**
     * 
     * @param msg
     */
    public static void throwNewException(String msg) throws Exception {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        Validator v = new Validator();
        v.addError(msg);
        v.throwExceptionIfErrors();
    }

//    /**
//     * 
//     * @param msg
//     */
//    public static void throwNewDontExpectToGetHereException() throws Exception {
//        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
//        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
//
//    	String msg = "Don't expect to get here";
//        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
//        Validator v = new Validator();
//        v.addError(msg);
//        v.throwExceptionIfErrors();
//    }

    /**
     * 
     * @param msg
     */
    public static void throwNewException(Exception ex) throws Exception {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewException";
        }
        DetailReport.recordFail(testCallerDetails, actualCallerDetails, ex.getLocalizedMessage (), ex.getLocalizedMessage ());
        Validator v = new Validator();
        v.addError(msg);
        v.throwExceptionIfErrors();
    }
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(String msg) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        Validator v = new Validator();
        v.addError(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
//    /**
//     * 
//     * @param msg
//     */
//    public static void throwNewDontExpectToGetHereRuntimeException() {
//        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
//        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
//
//    	String msg = "Don't expect to get here";
//        DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
//        Validator v = new Validator();
//        v.addError(msg);
//        v.throwRuntimeExceptionIfErrors();
//    }
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(Exception ex) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewRuntimeException";
        }
        DetailReport.recordFail(testCallerDetails, actualCallerDetails, ex.getLocalizedMessage (), ex.getLocalizedMessage ());
        Validator v = new Validator();
        v.addError(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
    /**
     * Formats an error message for surefire & DetailReport.
     * Includes debug info.
     * 
     * @return
     */
    private String getErrorMsgIfExists() {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        mergeDebugs();
        //TODO this is where I could add debug info to the detail report

        String msg = null;
        if (doErrorsExist()) {          
                StringBuffer sb = new StringBuffer();

                //
                // check if the client only wants limited amount of info for the surefire report
                if (ErrorOption.SUREFIRE_SHORT.equals(sErrorOption)) {
                    sb.append("\n<br>" + errors.get(0));
                    sb.append("\n... more ... see target/QATools.org-detailReport.html");
                } else { //show all errors
                    formatErrors();
                    for (String str : errors) {
                        sb.append(str);
                    }
                }
    
                
                if (doDebugsExist()) {
                    DetailReport.addDebugInfo(testCallerDetails, debugs);	// all debugs added to DetailReport
                    if (DebugOption.EVERYWHERE.equals(sDebugOption)) {		// client may opt to limit debug info that shows in surefire
                        formatDebugs();
                        for (String str : debugs) {
                            sb.append(str);
                        }
                    }
                }
                sb.append("\n");
                msg = sb.toString();
        }  // if doErrorsExist
        
        //
        // 
        DetailReport.autoGenerateDetailReport(testCallerDetails);
        return msg;
    }
    
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

    /**
     * Resets there error & debug lists
     */
    public void reset() {
        errors = null;
        debugs = null;
        //resetStaticProperties(); //not sure we want to reset these here
    }
    
    /**
     * Resets static debugs and 
     *         Validator.disableStaticDebugging();
     *         Validator.enableStaticDebuggingAutoCleanOnMerge();
     * 
     */
    public static void resetStaticProperties() {
        sDebugs = null;
        Validator.disableStaticDebugging();
        Validator.enableStaticDebuggingAutoCleanOnMerge();
    }
    
    
    // ------------------------------------------------------
    // ----- Re-implements these Junit APIs
    // ------------------------------------------------------

    
	/**
	 * Asserts that a condition is true. If it isn't it adds the error to the error list..
	 */
	public void assertTrue(String message, boolean condition) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (!condition) {
        	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected True but was False");
			addError(message + " expected True but was False");
		} else {
			DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected True and was True");
		}
	}
	

	
	/**
	 * Asserts that a condition is false. If it isn't it adds the error to the error list.
	 */
	public void assertFalse(String message, boolean condition) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (condition) {
			addError(message + " expected False but was True");
        	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected False but was True");
		} else {
			DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected False and was False");
		}
	}
	


	/**
	 * Adds the message to the error list
	 */
	public void fail(String message) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message);
		addError(message);
	}



	/**
	 * Asserts that two objects are equal. If they are not
	 * add the message to the error list
	 */
	public void assertEquals(String message, Object expected, Object actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected == null && actual == null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
	    	return;
		}
		if (expected != null && expected.equals(actual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
			return;
		}
		addError(formatErrorMsg(message, expected, actual));
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, expected, actual));
	}


	/**
	 * Asserts that two Strings are equal. 
	 */
	public void assertEquals(String message, String expected, String actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (cleanString(expected) == null && cleanString(actual) == null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + cleanStringForReport(expected) + "', actual:'" + cleanStringForReport(actual) + "'");
			return;
		}
		if (cleanString(expected) != null && cleanString(expected).equals(cleanString(actual))) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + cleanStringForReport(expected) + "', actual:'" + cleanStringForReport(actual) + "'");
			return;
		}
		addError(formatErrorMsg(message, cleanStringForReport(expected), cleanStringForReport(actual)));
    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, cleanStringForReport(expected), cleanStringForReport(actual)));
	}


	
	/**
	 * Asserts that two doubles are equal concerning a delta.  If the expected
	 * value is infinity then the delta value is ignored.
	 */
	public void assertEquals(String message, double expected, double actual, double delta) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		// handle infinity specially since subtracting to infinite values gives NaN and the
		// the following test fails
		if (Double.isInfinite(expected)) {
			if (!(expected == actual)) {
				DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Double(expected), new Double(actual)));
				addError(formatErrorMsg(message, new Double(expected), new Double(actual)));
			}
		} else if (!(Math.abs(expected-actual) <= delta)) { // Because comparison with NaN always returns false
			DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Double(expected), new Double(actual)));
			addError(formatErrorMsg(message, new Double(expected), new Double(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}
	


	/**
	 * Asserts that two floats are equal concerning a delta. If they are not
	 * an AssertionFailedError is thrown with the given message.  If the expected
	 * value is infinity then the delta value is ignored.
	 */
	public void assertEquals(String message, float expected, float actual, float delta) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

 		// handle infinity specially since subtracting to infinite values gives NaN and the
		// the following test fails
		if (Float.isInfinite(expected)) {
			if (!(expected == actual)) {
				addError(formatErrorMsg(message, new Float(expected), new Float(actual)));
				DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Float(expected), new Float(actual)));
			}
		} else if (!(Math.abs(expected-actual) <= delta)) {
			addError(formatErrorMsg(message, new Float(expected), new Float(actual)));			
			DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Float(expected), new Float(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

	
	/**
	 * Asserts that two longs are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, long expected, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Long(expected), new Long(actual)));
	    	addError(formatErrorMsg(message, new Long(expected), new Long(actual)));
		} else
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
	}
	
	/**
	 * Asserts that two booleans are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, boolean expected, boolean actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Boolean(expected), new Boolean(actual)));
	    	addError(formatErrorMsg(message, new Boolean(expected), new Boolean(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}
	
	/**
	 * Asserts that two bytes are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, byte expected, byte actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Byte(expected), new Byte(actual)));
	    	addError(formatErrorMsg(message, new Byte(expected), new Byte(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

  	
	/**
	 * Asserts that two chars are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, char expected, char actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError(formatErrorMsg(message, new Character(expected), new Character(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Character(expected), new Character(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}

	/**
	 * Asserts that two shorts are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertEquals(String message, short expected, short actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError(formatErrorMsg(message, new Short(expected), new Short(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Short(expected), new Short(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
	}

	/**
	 * Asserts that two ints are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertEquals(String message, int expected, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (expected != actual) {
			addError(formatErrorMsg(message, new Integer(expected), new Integer(actual)));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, formatErrorMsg(message, new Integer(expected), new Integer(actual)));
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected:'" + expected + "', actual:'" + actual + "'");
		}
  	}

	/**
	 * Asserts that an object isn't null. 
	 */
	public void assertNotNull(String message, Object object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (object == null) {
			addError(message + " expected not null, but was null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected not null, but was null");
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected not null, and was not null");
		}
	}
	
	/**
	 * Asserts that an object is null. 
	 */
	public void assertNull(String message, Object object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

		if (object != null) {
			addError(message + " expected null, but was not null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected null, but was not null");
		} else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected null, and was null");
		}
	}

	   /**
     * Asserts that an object is null. 
     */
    public void assertNull(String message, String object) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (cleanString(object) != null) {
            addError(message + " expected null, but was not null - " + cleanStringForReport(object));
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected null, but was not null - " + cleanStringForReport(object));
        } else {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected null, and was null");
        }
    }

	
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * asserts list is not null and not empty
	 * @param list
	 */
	public <T extends Object> void assertListNotNullNotEmpty(String message, List<T> list) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (list == null) {
			addError(message + " expected notNullNotEmpty, but was null");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was null");
        } else if (list.size() == 0) {
			addError(message + " expected notNullNotEmpty, but was empty");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was empty");
        } else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected notNullNotEmpty, and was notNullNotEmpty");
        }        
	}
	
	/**
	 * asserts list is not null and not empty
	 * @param list
	 */
	public void assertStrNotNullNotEmpty(String message, String str) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (cleanString(str) == null) {
			addError(message + " expected notNullNotEmpty, but was null - " + cleanStringForReport(str));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was null - " + cleanStringForReport(str));
        } else if (cleanString(str).length() == 0) {
			addError(message + " expected notNullNotEmpty, but was empty - " + cleanStringForReport(str));
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected notNullNotEmpty, but was empty - " + cleanStringForReport(str));
        } else {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected notNullNotEmpty, and was " + cleanStringForReport(str));
        }        
	}
	
    // ------------------------------------------------------
    // ----- Greater Than
    // ------------------------------------------------------

	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThan(String message, long smallExpected, long bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected < bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " < " + bigActual);
       } else {
			addError(message + " " + smallExpected + " NOT < " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT < " + bigActual);
       }
	}
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThan(String message, int smallExpected, int bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected < bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " < " + bigActual);
       } else {
			addError(message + " " + smallExpected + " NOT < " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT < " + bigActual);
       }
	}

	
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThanEquals(String message, long smallExpected, long bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected <= bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " <= " + bigActual);
       } else {
			addError(message + " " + smallExpected + " NOT <= " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT <= " + bigActual);
       }
	}
	/**
	 * actual should be greater than expected
	 * 
	 * @param message
	 * @param smallExpected
	 * @param bigActual
	 */
	public void assertGreaterThanEquals(String message, int smallExpected, int bigActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (smallExpected <= bigActual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + smallExpected + " <= " + bigActual);
       } else {
			addError(message + " " + smallExpected + " NOT <= " + bigActual);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + smallExpected + " NOT <= " + bigActual);
       }
	}

    // ------------------------------------------------------
    // ----- Less Than
    // ------------------------------------------------------

    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThan(String message, long bigExpected, long smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected > smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " > " + smallActual);
       } else {
            addError(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT > " + smallActual);
       }
    }
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThan(String message, int bigExpected, int smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected > smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " > " + smallActual);
       } else {
            addError(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT > " + smallActual);
       }
    }

    
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThanEquals(String message, long bigExpected, long smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected >= smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " >= " + smallActual);
       } else {
            addError(message + " " + bigExpected + " NOT >= " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT >= " + smallActual);
       }
    }
    /**
     * actual should be greater than expected
     * 
     * @param message
     * @param bigExpected
     * @param smallActual
     */
    public void assertLessThanEquals(String message, int bigExpected, int smallActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

       if (bigExpected >= smallActual) {
            DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " " + bigExpected + " >= " + smallActual);
       } else {
            addError(message + " " + bigExpected + " NOT > " + smallActual);
            DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " " + bigExpected + " NOT >= " + smallActual);
       }
    }
    
    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * The actual list should contain the expected object
	 * 
	 * @param expected
	 * @param actual
	 */
	public <T extends Object> void assertContains(String message, T expected, List<T> actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (actual==null) {
			addError(message + " actual is NULL");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual is NULL");
        	
        } else if (actual.size() == 0) {
			addError(message + " actual size is zero");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual size is zero");
        	
        } else if (actual.contains(expected)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + actual.toString() + "' contains expected '" + expected + "'");
        } else {
			addError(message + " actual '" + actual.toString() + "' does NOT contain expected '" + expected + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual '" + actual.toString() + "' does NOT contain expected '" + expected + "'");
        }
	}
	
	/**
	 * The actual list should contain the expected object
	 * 
	 * @param expected
	 * @param actual
	 */
	public <T extends Object> void assertContains(String message, List<T> expected, T actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        if (expected==null) {
			addError(message + " expected is NULL");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected is NULL");
        	
        } else if (expected.size() == 0) {
			addError(message + " expected size is zero");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected size is zero");

        } else if (expected.contains(actual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + expected.toString() + "' contains actual '" + actual + "'");
        } else {
			addError(message + " expected '" + expected.toString() + "' does NOT contain actual '" + actual + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected '" + expected.toString() + "' does NOT contain actual '" + actual + "'");
        }
	}

	/**
	 * The actual string must be contained within the expected
	 * 
	 * @param message
	 * @param longExpected
	 * @param shortActual
	 */
	public void assertStrContainsStr(String message, String longExpected, String shortActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String tmpLongExpected = cleanString(longExpected);
        String tmpShortActual = cleanString(shortActual);
        
        if (tmpLongExpected==null && tmpShortActual==null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + cleanStringForReport(longExpected) + "' contains actual '" + cleanStringForReport(shortActual) + "'");
        	
        } else if (tmpLongExpected.contains(tmpShortActual)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " expected '" + cleanStringForReport(longExpected) + "' contains actual '" + cleanStringForReport(shortActual) + "'");
        } else {
			addError(message + " expected '" + longExpected + "' does NOT contain actual '" + shortActual + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " expected '" + cleanStringForReport(longExpected) + "' does NOT contain actual '" + cleanStringForReport(shortActual) + "'");
        }
	}

	/**
	 * The expected must be contained within the actual
	 * 
	 * @param message
	 * @param shortExpected
	 * @param longActual
	 */
	public void assertStrContainsStr2(String message, String shortExpected, String longActual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String tmpShortExpected = cleanString(shortExpected);
        String tmpLongActual = cleanString(longActual);
         
        if (tmpLongActual==null && tmpShortExpected==null) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + cleanStringForReport(longActual) + "' contains expected '" + cleanStringForReport(shortExpected) + "'");
        	
        } else if (tmpLongActual.contains(tmpShortExpected)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, message + " actual '" + cleanStringForReport(longActual) + "' contains expected '" + cleanStringForReport(shortExpected) + "'");
        } else {
			addError(message + " actual '" + cleanStringForReport(longActual) + "' does NOT contain expected '" + cleanStringForReport(shortExpected) + "'");
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, message + " actual '" + cleanStringForReport(longActual) + "' does NOT contain expected '" + cleanStringForReport(shortExpected) + "'");
        }
	}

	// ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	/**
	 * 
	 * @param message
	 * @param list
	 * @param list2
	 */
	public void assertEquals(String message, List<Object> expected, List<Object> actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
		
        if (expected==null && actual!=null) {
        	String msg = message + " expected is null, actual is not null(size:" + actual.size() + ")";
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        	return;
        	
        } else if (expected!=null && actual==null) {
        	String msg = message + " expected is not null(size:" + expected.size() + ") actual is null)";
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
        	return;        	
        } 

        if (expected.size() != actual.size()) {
        	String msg = message + " expected.size=" + expected.size() + ", actual.size=" + actual.size();
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
	    	addDebug(message + ", expected=" + expected.toString());
	    	addDebug(message + ", actual=" + actual.toString());
        	return;        	        	
        }
        
        boolean flag = true;
        for (int i=0; i<expected.size(); i++) {
        	Object e = expected.get(i);
        	Object a = actual.get(i);
        	if (!e.equals(a)) {
        		flag = false;
        	}
        }
        if (flag) {
        	String msg = message + ", expected.list matches actual.list";
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
        } else {
        	String msg = message + ", expected.list does NOT match actual.list";
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, msg, msg);
	    	addDebug(message + ", expected=" + expected.toString());
	    	addDebug(message + ", actual=" + actual.toString());        	
        }
        
	}
	
	// ------------------------------------------------------
    // ----- assertNotEquals
    // ------------------------------------------------------

	/**
	 * Asserts that two objects are NOT equal. If they are equal
	 * add the message to the error list
	 */
	public void assertNotEquals(String message, Object expected, Object actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";
		if (expected == null && actual == null) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, "NOT expected:'" + expected + "', actual:'" + actual + "'");
			addError(msg);
	    	return;
		}
		if (expected != null && expected.equals(actual)) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, "NOT expected:'" + expected + "', actual:'" + actual + "'");
			addError(msg);
			return;
		}
    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
	}


	/**
	 * Asserts that two Strings are NOT equal. 
	 */
	public void assertNotEquals(String message, String expected, String actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String ex = cleanString(expected);
        String a = cleanString(actual);
        String msg = message + " - NOT expected:'" + ex + "', actual:'" + a + "'";

		if (ex == null && a == null) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
			addError(msg);
			return;
		}
		if (ex != null && ex.equals(a)) {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
			addError(msg);
			return;
		}
    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
	}

	
	/**
	 * Asserts that two longs are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertNotEquals(String message, long expected, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();

        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";
        
		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
    		addError(msg);
		}
	}
	
	/**
	 * Asserts that two booleans are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
	public void assertNotEquals(String message, boolean expected, boolean actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";

		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
	    	addError(msg);
		}
  	}
	

	/**
	 * Asserts that two ints are equal. If they are not
	 * an AssertionFailedError is thrown with the given message.
	 */
  	public void assertNotEquals(String message, int expected, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - NOT expected:'" + expected + "', actual:'" + actual + "'";

		if (expected != actual) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
  	}
	
	// ------------------------------------------------------
    // ----- range
    // ------------------------------------------------------

  	/**
  	 * 
  	 * @param message
  	 * @param expectedLow
  	 * @param expectedHigh
  	 * @param actual
  	 */
	public void assertRange(String message, int expectedLow, int expectedHigh, int actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - expected range:'" + expectedLow + "-" + expectedHigh + "', actual:'" + actual + "'";

		if ((expectedLow <= actual) && (actual <= expectedHigh)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
	}

  	/**
  	 * 
  	 * @param message
  	 * @param expectedLow
  	 * @param expectedHigh
  	 * @param actual
  	 */
	public void assertRange(String message, long expectedLow, long expectedHigh, long actual) {
        String[] testCallerDetails = StackUnwinder.getCallerTestMethodDetails();
        String[] actualCallerDetails = StackUnwinder.getActualCallerDetails();
        String msg = message + " - expected range:'" + expectedLow + "-" + expectedHigh + "', actual:'" + actual + "'";

		if ((expectedLow <= actual) && (actual <= expectedHigh)) {
	    	DetailReport.recordPass(testCallerDetails, actualCallerDetails, msg);
		} else {
			addError(msg);
	    	DetailReport.recordFail(testCallerDetails, actualCallerDetails, message, msg);
		}
	}

	// ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	
	String formatPassMsg(String message, Object expected, Object actual) {
		String formatted= "";
		if (message != null)
			formatted= message+" ";
		return formatted+"expected:'"+expected+"' and was:'"+actual+"'";
	}
	String formatErrorMsg(String message, Object expected, Object actual) {
		String formatted= "";
		if (message != null)
			formatted= message+" ";
		return formatted+"expected:'"+expected+"' but was:'"+actual+"'";
	}


    // ------------------------------------------------------
    // ----- StringProcessingOption
    // ------------------------------------------------------


	
	/**
	 * 		StringProcessingOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringProcessingOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringProcessingOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringProcessingOption
	 */
	public static void setStringCleaningOption(StringCleaningOption stringProcessingOption) {
	    //System.out.println("\nEnter setStringProcessingOption " + stringProcessingOption + "\n");
		
		if (stringProcessingOption == null) {
		    //System.out.println ("stringProcessingOption == null");
			setStringCleaningOptions(null);
		}
		setStringCleaningOptions(Arrays.asList(stringProcessingOption));
	}
	
	/**
	 * 		StringCleaningOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringCleaningOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringCleaningOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringCleaningOptions
	 */
	public static void setStringCleaningOptions(List<StringCleaningOption> stringCleaningOptions) {
	    if (stringCleaningOptions==null || 0 == stringCleaningOptions.size ()) {
	        sStringCleaningOptions = null;
	        //System.out.println("sStringProcessingOptions = null;");
	    } else {
            sStringCleaningOptions = new LinkedList<StringCleaningOption>();
            for (StringCleaningOption option : stringCleaningOptions) {
                sStringCleaningOptions.add (option);
                //System.out.println("adding " + option);
            }
	    }

	}
	
	
	/**
	 * 		StringCleaningOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringCleaningOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringCleaningOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringProcessingOptions
	 */
    public static void addStringCleaningOption(StringCleaningOption stringCleaningOption) {
        if (sStringCleaningOptions == null) {
            sStringCleaningOptions = new LinkedList<StringCleaningOption>();
        }
        sStringCleaningOptions.add (stringCleaningOption);
    }

	
	/**
	 * Sets values to be used for StringProcessingOption.NULL_EQUIVALENCE. 
	 * Automatically enables StringProcessingOption.NULL_EQUIVALENCE if not already enabled
	 *  
	 * For example, "", " ", "{}", "null", "0.0" can be set equivalent to null so that the Validator treats your values as equivalent to null
	 */
	public static void setNullEquivalenceString(String str) {
		if (str == null) {
			setNullEquivalenceStrings(null);
		}
		setNullEquivalenceStrings(Arrays.asList(str));
	}

	/**
	 * Sets values to be used for StringProcessingOption.NULL_EQUIVALENCE. 
	 * Automatically enables StringProcessingOption.NULL_EQUIVALENCE if not already enabled
	 *  
	 * For example, "", " ", "{}", "null", "0.0" can be set equivalent to null so that the Validator treats your values as equivalent to null
	 */
	public static void setNullEquivalenceStrings(List<String> strings) {
	    if (strings==null || 0 == strings.size ()) {
	        sNullEquivalenceStrings = null;
	    } else {
            addStringCleaningOption(StringCleaningOption.NULL_EQUIVALENCE);   // this is also required to do null equiv processing          
	        sNullEquivalenceStrings = new LinkedList<String>();
	        for (String s : strings) {
	            //System.out.println("setNullEquivalenceStrings add - " + s);
	        	if (s != null) {
		            sNullEquivalenceStrings.add (s);
	        	}
	        }
	    }
	}

    // ------------------------------------------------------
    // ----- methods for processing (cleaning up) strings
    // ------------------------------------------------------

	/**
	 * Returns Object[]
	 * 		Boolean.TRUE / FALSE if the string was changed
	 * 		original value
	 * 		new clean value
	 * @param str
	 * @return
	 */
	private static Object[] cleanStringBase(String str) {
		Object[] result = new Object[3];
		result[0] = Boolean.FALSE;	// assume string does not need cleaning
		result[1] = str;			// this element contains the original value
		result[2] = null;			//doesn't matter unless processing occurs
		
		if (str == null) {
	        //System.out.println("processString returning input = " + str);
			return result;
		}

        //System.out.println("processString input = '" + str + "'");

        if (sStringCleaningOptions==null) {
	        //System.out.println("sStringProcessingOptions==null");
			return result;
		} 
        
        
		//
        // trim string
		String s = str;
		if (sStringCleaningOptions.contains(StringCleaningOption.TRIM)) {
            //System.out.println("processString in trim(1) s = '" + s + "'");
			s = s.trim();
            //System.out.println("processString in trim(2) s = '" + s + "'");
		}
		
		
		//
		// remove double spaces
		if (sStringCleaningOptions.contains(StringCleaningOption.REMOVE_DUPLICATE_SPACES)) {
		    String SINGLE_SPACE = " ";
		    String DOUBLE_SAPCE = "  ";
			while(s.contains(DOUBLE_SAPCE)) {
				s = s.replaceAll(DOUBLE_SAPCE, SINGLE_SPACE);
			}
		}
		
		
		//
		// Null equivalence changes these
		//		"" to null
		//		" " to null
		//		"0.0" to null
		if (sStringCleaningOptions.contains(StringCleaningOption.NULL_EQUIVALENCE)) {
		    //System.out.println("Yes - NULL_EQUIVALENCE");
		    if (sNullEquivalenceStrings!=null) {
		        //System.out.println("Yes - sNullEquivalenceStrings!=null");
		        if (0<sNullEquivalenceStrings.size()) {
		            //System.out.println("Yes - 0<sNullEquivalenceStrings.size()");
		            for (String b : sNullEquivalenceStrings) {
		                //System.out.println("null equiv setting = " + b + ", s = " + s);
		                if (b!=null && b.equals(s)) {
		                    s = null;
		                    //System.out.println("set s = null");
		                } // 
		            } //for
		        } //if 0<sNullEquivalenceStrings.size()
		    } // if sNullEquivalenceStrings!=null
		} // if NULL_EQUIVALENCE
		
		if (!str.equals(s)) {
			result[0] = Boolean.TRUE;
			result[2] = s;
		}
		
		return result;		
	}

	/**
	 * 
	 * @param str
	 * @return
	 */
	private static String cleanString(String str) {
		Object[] result = cleanStringBase(str);
		if (Boolean.TRUE.equals(result[0])) {
			return (String)result[2];	//the modified value
		}
		return (String) result[1];	//the original value
	}

	/**
	 * Convenience method for better reporting. The detail report will show the original & new value of any strings that are processed.
	 * 
	 * If processing modified the original string, this returns
	 * 		foo(  foo  )
	 * otherwise returns
	 * 		foo
	 * 
	 * @param str
	 * @return
	 */
	private static String cleanStringForReport(String str) {
		if (sStringCleaningForReportFlag == false) {
			return str;
		}
		Object[] result = cleanStringBase(str);
		if (Boolean.FALSE.equals(result[0])) {
			return (String)result[1]; //the original value
		}
		return (String)(result[2] + "(" + result[1] + ")"); // foo(   foo   )
	}

	
    // ------------------------------------------------------
    // ----- Print Format
    // ------------------------------------------------------


    /**
     * Allows you to set the print format suitable for your environment
     *
     *     	PrintFormat.NEWLINE - errors show up on a separate line
     *     
     *     	PrintFormat.BR - html formatted <br> is added to each error for viewing in a browser
     *     
     *    	PrintFormat.NUMBER_ERRORS - number errors
     *    
     *    	PrintFormat.HASH_MARKS - section headers for errors and debugs
     * 
     * @param printFormats
     */
    public static void setPrintFormats(List<PrintFormat> printFormats) {
    	if (printFormats == null || printFormats.size() == 0) {
    		sPrintFormats = null;
    		return;
    	}
    	
    	//
    	// The reason we don't just do this assignment sPrintFormats = printFormats
    	// is Arrays.asList(....)  creates an immutable list
    	sPrintFormats = null;	// over -rides the default settings
    	for (PrintFormat pf : printFormats) {
    		addPrintFormat(pf);
    	}
	}

    /**
     * Allows you to set the print format suitable for your environment
     *
     *     	PrintFormat.NEWLINE - errors show up on a separate line
     *     
     *     	PrintFormat.BR - html formatted <br> is added to each error for viewing in a browser
     *     
     *    	PrintFormat.NUMBER_ERRORS - number errors
     *    
     *    	PrintFormat.HASH_MARKS - section headers for errors and debugs
     * 
     * @param printFormats
     */
    public static void addPrintFormat(PrintFormat printFormat) {
    	if (sPrintFormats == null || sPrintFormats.size()==0) {
    		sPrintFormats = new LinkedList<PrintFormat>();
    	}
    	sPrintFormats.add(printFormat);
	}
    
    /**
     *     	ErrorOption.SUREFIRE_FULL - configure report to show all details in your surefire reports
     *     
     *     	ErrorOption.SUREFIRE_SHORT - only show the first error in your surefire reports. Get the full detail list in target/QATools.org-detailReport.html

     * 
     * @param printFormat
     */
    public static void setErrorOption(ErrorOption errorOption) {
    	sErrorOption = errorOption;
	}

    /**
     *     DebugOption.EVERYWHERE - debug details show up in surefire reports and QATools reports
     *    
     *     DebugOption.QATOOLS_DETAIL_REPORT_ONLY - only show debug details in QATools reports. 
     *     Debug messages do not show up in surefire reports.
     * 
     * 
     * @param printFormat
     */
    public static void setDebugOption(DebugOption debugOption) {
    	sDebugOption = debugOption;
	}

    
    /**
     * @return
     */
    private boolean isNewline() {
    	if (sPrintFormats.contains(PrintFormat.NEWLINE)) {
    		return true;
    	}
    	return false;
    }

    /**
     * @return
     */
    private boolean isBR() {
    	if (sPrintFormats.contains(PrintFormat.BR)) {
    		return true;
    	}
    	return false;
    }

    /**
     * @return
     */
    private boolean isNumberErrors() {
        if (sPrintFormats.contains(PrintFormat.NUMBER_ERRORS)) {
            return true;
        }
        return false;
    }

    /**
     * @return
     */
    private boolean isHash() {
        if (sPrintFormats.contains(PrintFormat.HASH_MARKS)) {
            return true;
        }
        return false;
    }

    /**
     * 
     */
    private void formatErrors() {
    	if (!doErrorsExist()) {
    		return;
    	}
    	String br = "";
    	String newline = "";
    	for (int i=0; i<errors.size(); i++) {
    		String str = errors.get(i);
    		int b = i + 1;
    		
    		if (isNumberErrors()) {
    			str = "(" + b + ") " + str;
    		}	//isNumberErrors

    		if (isBR()) {
    			br = "<br>";
    			str = "<br>" + str;
    		}	//isBR

    		if (isNewline()) {
    			newline = "\n";
    			str = "\n" + str;
    		}	//isNewline

    		errors.set(i, str);
    	}	//for
    	
    	//
    	//only need this extra header info if there is also debug info
    	if (doDebugsExist()) {
    	    String hashMarks = (isHash() ? newline + br + "##################################" : "");
        	errors.add(0, hashMarks);
        	errors.add(0, newline + br + "### Error details");
            errors.add(0, hashMarks);
    	}

    }	//method

    /**
     * 
     */
    private void formatDebugs() {    	
    	if (!doDebugsExist()) {
    		return;
    	}
    	String br="";
    	String newline = "";
    	for (int i=0; i<debugs.size(); i++) {
    		String str = debugs.get(i);
    		int b = i + 1;
    		
    		if (isNumberErrors()) {
    			str = "(" + b + ") " + str;
    		}	//isNumberErrors

    		if (isBR()) {
    			str = "<br>" + str;
    			br = "<br>";
    		}	//isBR

    		if (isNewline()) {
    			str = "\n" + str;
    			newline = "\n";
    		}	//isNewline

    		debugs.set(i, str);
    	}	//for
        String hashMarks = (isHash() ? newline + br + "##################################" : "");
        debugs.add(0, hashMarks);
		debugs.add(0, newline + br + "### Debug / Contextual Information");
		debugs.add(0, hashMarks);

    }	//method

    
	/**
	 * <pre>
	 * Enum class for different print formats
	 * 
	 * The order in which formatting occurs is:
	 * 
	 * 			NUMBER_ERRORS - numbers each error (1), (2) etc
	 * 			BR	- adds a "<br>" to each error so they format nicely in HTML
	 * 			NEWLINE - adds a newline "\n" char to each error so they print nicely
	 *         HASH_MARKS - wraps error & debug messages with hash marks for visual separation of errors & debugs
	 * 
	 * </pre>
	 * 
	 * @author Scott Wilson - QATools.org
	 *
	 */
	public enum PrintFormat {
		NUMBER_ERRORS,
		NEWLINE,
		BR,
		HASH_MARKS
	}

	/**
	 * <pre>
	 * Enum class for different debug options
	 * 
	 * 			QATOOLS_DETAIL_REPORT_ONLY - Debug info is ONLY displayed on the QATools Detail Report.
	 * 					This is useful if you want to disable excessive debug details from showing up on your continuous build system.
	 * 			EVERYWHERE	- This is the default. All debug info shows up in all reports, QATools, surefure, junit, etc
	 * 
	 * </pre>
	 * 
	 * @author Scott Wilson - QATools.org
	 *
	 */
	public enum DebugOption {
		QATOOLS_DETAIL_REPORT_ONLY,
		EVERYWHERE
	}
	

	/**
	 * <pre>
	 * Enum class for different error options
	 * 
	 * 			SUREFIRE_SHORT - Shorten the error details in the surefire / junit reports. The formatting may be whacky in some reports.
	 * 				The QATools detail report will always include all details regardless of this setting.
	 * 				It's just a way to control how much detail shows up in your continuous build reports.
	 * 			SUREFIRE_FULL	- This is the default. Full error details show up in surefire / junit reports.
	 * 
	 * </pre>
	 * 
	 * @author Scott Wilson - QATools.org
	 *
	 */
	public enum ErrorOption {
		SUREFIRE_SHORT,
		SUREFIRE_FULL
	}

	/**
	 * <pre>
	 * Enum class for preprocessing string inputs
	 * 		TRIM - "foo" = "foo", " foo", " foo ", "     foo     "
	 * 				- "" = " ", "     "
	 * 		REMOVE_DUPLICATE_SPACES - "foo bar car" = "foo   bar     car"
	 * 		NULL_EQUIVALENCE - if this option is configured, you can add null equivalence strings
	 * 			by invoking 
	 * 				addNullEquivalence("");
	 * 				addNullEquivalence(" ");
	 * 				addNullEquivalence("()");
	 * 				addNullEquivalence("[]");
	 * </pre>
	 * 
	 * @author Scott Wilson - QATools.org
	 *
	 */
	public enum StringCleaningOption {
		TRIM,
		REMOVE_DUPLICATE_SPACES,
		NULL_EQUIVALENCE
	}
	
	
	
	
}
